/*
 * @(#)AbstractCellView.java	1.0 03-JUL-04
 * 
 * Copyright (c) 2001-2004 Gaudenz Alder
 *  
 */
package org.jgraph.graph;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Stack;

import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeNode;

import org.jgraph.JGraph;

/**
 * The abstract base class for all cell views.
 * 
 * @version 1.0 1/3/02
 * @author Gaudenz Alder
 */

public abstract class AbstractCellView implements CellView, Serializable {

	/** Editor for the cell. */
	protected static transient GraphCellEditor cellEditor = new DefaultGraphCellEditor();

	/** Reference to the cell for this view */
	protected Object cell = null;

	/** Cached parent view */
	protected transient CellView parent = null;

	/** Cached child views. */
	protected transient java.util.List childViews = new ArrayList();

	/** Cached bounds of all children if vertex is a group */
	protected transient Rectangle2D groupBounds = DefaultGraphCell.defaultBounds;

	/**
	 * Hashtable for attributes. Initially created by calling
	 * <code>createLocalAttributes</code>. Value in this map override the
	 * values in <code>allAttributes</code>.
	 */
	protected AttributeMap attributes = new AttributeMap();

	/**
	 * Contains the complete set of attributes, ie. including the cell's
	 * attributes. The values in this map are overriden by the corresponding
	 * values in <code>attributes</code>.
	 */
	protected transient AttributeMap allAttributes = createFallbackAttributeMap();

	/**
	 * Constructs an empty abstract cell view. You should set a cell on this
	 * view using setCell before doing anything. Optionally you can also
	 * set a different attribute map using setAttributeMap.
	 * Note: To change the attribute map you should now use the 
	 * changeAttributes method.
	 * 
	 * @param cell
	 *            reference to the model object
	 */
	public AbstractCellView() {
	}
	
	/**
	 * Constructs a view for the specified model object, and invokes update on
	 * the new instance.
	 * 
	 * @param cell
	 *            reference to the model object
	 */
	public AbstractCellView(Object cell) {
		setCell(cell);
	}

	/**
	 * Returns the model object that this view represents.
	 */
	public Object getCell() {
		return cell;
	}

	/**
	 * Returns the model object that this view represents.
	 */
	public void setCell(Object cell) {
		this.cell = cell;
	}
	
	/**
	 * Create child views and reload properties. Invokes update first.
	 */
	public void refresh(GraphModel model, CellMapper mapper,
			boolean createDependentViews) {
		// Re-read global attributes
		allAttributes = model.getAttributes(cell);
		// Cache Cell Attributes in View
		update();
		// Cache Parent View
		if (mapper != null && model != null) {
			// Create parent only if it's visible in the graph
			Object par = model.getParent(cell);
			CellView tmp = mapper.getMapping(par, createDependentViews);
			if (tmp != parent)
				removeFromParent();
			parent = tmp;
		}
		// Re-load Child Views
		childViews.clear();
		for (int i = 0; i < model.getChildCount(cell); i++) {
			Object child = model.getChild(cell, i);
			CellView view = mapper.getMapping(child, createDependentViews);
			if (!model.isPort(child) && view != null)
				childViews.add(view);
		}
	}

	/**
	 * Update attributes and recurse children.
	 */
	public void update() {
		updateAllAttributes();
		// Notify Parent
		groupBounds = null;
		childUpdated();
	}

	/**
	 * This method implements the merge between the cell's and the view's
	 * attributes. The view's attributes override the cell's attributes.
	 */
	protected void updateAllAttributes() {
		if (allAttributes != null) {
			allAttributes = (AttributeMap) allAttributes.clone();
		} else
			allAttributes = createFallbackAttributeMap();
		// Override global with local attributes
		allAttributes.putAll(attributes);
	}

	/**
	 * Default implementation returns an <code>AttributeMap</code>
	 * 
	 * @return a new <code>AttributeMap</code>
	 */
	protected AttributeMap createFallbackAttributeMap() {
		return new AttributeMap();
	}

	public void childUpdated() {
		if (parent != null)
			parent.childUpdated();
		groupBounds = null;
	}

	//
	// Graph Structure
	//

	/**
	 * Returns the parent view for this view.
	 */
	public CellView getParentView() {
		return parent;
	}

	/**
	 * Returns the child views of this view.
	 */
	public CellView[] getChildViews() {
		CellView[] array = new CellView[childViews.size()];
		childViews.toArray(array);
		return array;
	}

	/**
	 * Returns all views, including descendants that have a parent in
	 * <code>views</code> without the PortViews. Note: Iterative
	 * Implementation using view.getChildViews. This returns the array in
	 * inverse order, ie with the top most cell view at index 0.
	 */
	public static CellView[] getDescendantViews(CellView[] views) {
		Stack stack = new Stack();
		for (int i = 0; i < views.length; i++)
			stack.add(views[i]);
		ArrayList result = new ArrayList();
		while (!stack.isEmpty()) {
			CellView tmp = (CellView) stack.pop();
			Object[] children = tmp.getChildViews();
			for (int i = 0; i < children.length; i++)
				stack.add(children[i]);
			result.add(tmp);
		}
		CellView[] ret = new CellView[result.size()];
		result.toArray(ret);
		return ret;
	}

	/**
	 * Removes this view from the list of childs of the parent.
	 */
	public void removeFromParent() {
		if (parent instanceof AbstractCellView) {
			java.util.List list = ((AbstractCellView) parent).childViews;
			list.remove(this);
		}
	}

	/**
	 * Returns true if the view is a leaf.
	 */
	public boolean isLeaf() {
		return childViews.isEmpty();
	}

	//
	// View Attributes
	//

	/**
	 * Return the attributes of the view.
	 */
	public AttributeMap getAttributes() {
		return attributes;
	}

	/**
	 * Return the attributes of the view.
	 */
	public void setAttributes(AttributeMap attributes) {
		this.attributes = attributes;
	}

	/**
	 * Returns the attributes of the view combined with the attributes of the
	 * corresponding cell. The view's attributes override the cell's attributes
	 * with the same key.
	 */
	public AttributeMap getAllAttributes() {
		return allAttributes;
	}

	/**
	 * Applies <code>change</code> to the attributes of the view and calls
	 * update.
	 */
	public Map changeAttributes(Map change) {
		Map undo = attributes.applyMap(change);
		update();
		return undo;
	}

	//
	// View Methods
	//

	/**
	 * Returns the cached bounds for the group if isleaf is false
	 */
	public Rectangle2D getBounds() {
		if (!isLeaf()) {
			if (groupBounds == null)
				updateGroupBounds();
			return groupBounds;
		}
		return null;
	}

	/**
	 * Returns the bounding box for the specified views.
	 */
	public static Rectangle2D getBounds(CellView[] views) {
		if (views != null && views.length > 0) {
			Rectangle2D ret = null;
			for (int i = 0; i < views.length; i++) {
				if (views[i] != null) {
					Rectangle2D r = views[i].getBounds();
					if (r != null) {
						if (ret == null)
							ret = views[i].getAttributes().createRect(r);
						else
							Rectangle2D.union(ret, r, ret);
					}
				}
			}
			return ret;
		}
		return null;
	}

	/**
	 * Sets the bounds of <code>view</code>. Calls translateView and
	 * scaleView.
	 */
	public void setBounds(Rectangle2D bounds) {
		Rectangle2D oldBounds = getBounds();
		Dimension last;
		if (oldBounds == null)
			oldBounds = getAttributes().createRect();
		Point2D p0 = getAttributes().createPoint(oldBounds.getX(),
				oldBounds.getY());
		Point2D pe = getAttributes().createPoint(bounds.getX(), bounds.getY());
		Rectangle2D localBounds = getAttributes().createRect(bounds);
		if (GraphConstants.isMoveable(getAttributes()) && !pe.equals(p0))
			translate(pe.getX() - p0.getX(), pe.getY() - p0.getY());
		else
			localBounds.setFrame(localBounds.getX(), localBounds.getY(), bounds
					.getWidth()
					- pe.getX() + p0.getX(), bounds.getHeight() - pe.getY()
					+ p0.getY());
		double lbw = localBounds.getWidth(), lbh = localBounds.getHeight();
		double obw = oldBounds.getWidth(), obh = oldBounds.getHeight();
		if ((lbw != obw || lbh != obh) && obw > 0 && obh > 0)
			scale(lbw / obw, lbh / obh, pe);
	}

	protected void updateGroupBounds() {
		// Note: Prevent infinite recursion by removing
		// child edges that point to their parent.
		CellView[] childViews = getChildViews();
		LinkedList result = new LinkedList();
		for (int i = 0; i < childViews.length; i++)
			if (includeInGroupBounds(childViews[i]))
				result.add(childViews[i]);
		childViews = new CellView[result.size()];
		result.toArray(childViews);
		Rectangle2D r = getBounds(childViews);
		if (GraphConstants.isOpaque(getAllAttributes())) {
			int groupBorder = GraphConstants.getGroupBorder(getAllAttributes());
			r.setFrame(r.getX() - groupBorder, r.getY() - groupBorder, r
					.getWidth()
					+ 2 * groupBorder, r.getHeight() + 2 * groupBorder);
		}
		groupBounds = r;
	}

	// This is used to exclude certain cell views from the group bounds
	// computation. This implementation returns false
	// for edges that connect to one of their ancestor groups (eg. parent).
	protected boolean includeInGroupBounds(CellView view) {
		if (view instanceof EdgeView) {
			EdgeView edgeView = (EdgeView) view;
			if (edgeView.getCell() instanceof DefaultMutableTreeNode) {
				DefaultMutableTreeNode edge = (DefaultMutableTreeNode) edgeView
						.getCell();
				if (edgeView.getSource() != null
						&& edgeView.getSource().getParentView() != null) {
					Object src = edgeView.getSource().getParentView().getCell();
					if (src instanceof TreeNode) {
						TreeNode source = (TreeNode) src;
						if (((DefaultMutableTreeNode) source) // getParent()
								.isNodeDescendant(edge)) {
							return false;
						}
					}
				}
				if (edgeView.getTarget() != null
						&& edgeView.getTarget().getParentView() != null) {
					Object tgt = edgeView.getTarget().getParentView().getCell();
					if (tgt instanceof TreeNode) {
						TreeNode target = (TreeNode) tgt;
						if (((DefaultMutableTreeNode) target) // getParent()
								.isNodeDescendant(edge)) {
							return false;
						}
					}
				}
			}
		}
		return true;
	}

	/**
	 * Translates <code>view</code> (group) by <code>dx, dy</code>.
	 */
	protected void translate(double dx, double dy) {
		if (isLeaf())
			getAttributes().translate(dx, dy);
		else {
			Iterator it = childViews.iterator();
			while (it.hasNext()) {
				Object view = it.next();
				if (view instanceof AbstractCellView) {
					AbstractCellView child = (AbstractCellView) view;
					child.translate(dx, dy);
				}
			}
		}
	}

	/**
	 * Scale <code>view</code> (group) by <code>sx, sy</code>.
	 */
	protected void scale(double sx, double sy, Point2D origin) {
		if (isLeaf())
			getAttributes().scale(sx, sy, origin);
		else {
			Iterator it = childViews.iterator();
			while (it.hasNext()) {
				Object view = it.next();
				if (view instanceof AbstractCellView) {
					AbstractCellView child = (AbstractCellView) view;
					Map attributes = child.getAttributes();
					if (GraphConstants.isSizeable(attributes)
							|| GraphConstants.isAutoSize(attributes))
						child.scale(sx, sy, origin);
				}
			}
		}
	}

	/**
	 * Returns true if the view intersects the given rectangle.
	 */
	public boolean intersects(JGraph graph, Rectangle2D rect) {
		if (isLeaf() || GraphConstants.isOpaque(getAllAttributes())) {
			Rectangle2D bounds = getBounds();
			if (bounds != null)
				return bounds.intersects(rect);
		} else { // Check If Children Intersect?
			Iterator it = childViews.iterator();
			while (it.hasNext())
				if (((CellView) it.next()).intersects(graph, rect))
					return true;
		}
		return false;
	}

	//
	// View Editors
	//

	/**
	 * Returns a renderer component, configured for the view.
	 */
	public Component getRendererComponent(JGraph graph, boolean selected,
			boolean focus, boolean preview) {
		CellViewRenderer cvr = getRenderer();
		if (cvr != null)
			return cvr.getRendererComponent(graph, this, selected, focus,
					preview);
		return null;
	}

	public abstract CellViewRenderer getRenderer();

	/**
	 * Returns a cell handle for the view.
	 */
	public abstract CellHandle getHandle(GraphContext context);

	/**
	 * Returns a cell editor for the view.
	 */
	public GraphCellEditor getEditor() {
		return cellEditor;
	}

}
